/**
  Simple chess engine using smallchesslib.

  author: Miloslav Ciz
  license: CC0 1.0 (public domain)
           found at https://creativecommons.org/publicdomain/zero/1.0/
           + additional waiver of all IP

  This work's goal is to never be encumbered by any exclusive intellectual
  property rights. The work is therefore provided under CC0 1.0 + additional
  WAIVER OF ALL INTELLECTUAL PROPERTY RIGHTS that waives the rest of
  intellectual property rights not already waived by CC0 1.0. The WAIVER OF ALL
  INTELLECTUAL PROPERTY RGHTS is as follows:

  Each contributor to this work agrees that they waive any exclusive rights,
  including but not limited to copyright, patents, trademark, trade dress,
  industrial design, plant varieties and trade secrets, to any and all ideas,
  concepts, processes, discoveries, improvements and inventions conceived,
  discovered, made, designed, researched or developed by the contributor either
  solely or jointly with others, which relate to this work or result from this
  work. Should any waiver of such right be judged legally invalid or
  ineffective under applicable law, the contributor hereby grants to each
  affected person a royalty-free, non transferable, non sublicensable, non
  exclusive, irrevocable and unconditional license to this right.  
*/

#include <stdio.h>
#include <time.h>

#include "smallchesslib.h"

#define XBOARD_DEBUG 0

#if XBOARD_DEBUG
FILE *debugFile = 0;

char boardString[1024];
int boardStringPos = 0;

void putCharacterStr(char c)
{
  boardString[boardStringPos] = c;
  boardStringPos++;
}

void debugLogBoard(SCL_Board board)
{
  boardStringPos = 0;
  SCL_printBoardSimple(board,putCharacterStr,255,SCL_PRINT_FORMAT_UTF8);
  boardString[boardStringPos] = 0;
  fprintf(debugFile,"BOARD STATE:\n%s\n",boardString);
}
#endif

uint8_t *pixels = 0;

void putImagePixel(uint8_t pixel, uint16_t number)
{
  uint16_t index = number * 3;

  pixel = pixel ? 255 : 0;

  pixels[index] = pixel;
  pixels[index + 1] = pixel;
  pixels[index + 2] = pixel;
}

void putCharacter(char c)
{
  putchar(c);
}

void xboardSend(const char *string)
{
#if XBOARD_DEBUG
  fprintf(debugFile,"SENDING: '%s'\n",string);
  fflush(debugFile);
#endif

  printf("%s",string); 
  fflush(stdout);
}

uint8_t stringsEqual(const char *s1, const char *s2, int max)
{
  for (int i = 0; i < max; ++i)
  {
    if (*s1 != *s2)
      return 0;
 
    if (*s1 == 0)
      return 1;

    s1++;
    s2++;
  }

  return 1;
}
  
uint8_t paramPlayerW = 0;
uint8_t paramPlayerB = 0;
uint8_t paramBoard = 1;
uint8_t paramAnalyze = 255; // depth of analysis
uint8_t paramMoves = 0;
uint8_t paramXboard = 0;
uint8_t paramInfo = 1;
uint8_t paramDraw = 0;
uint8_t paramFlipBoard = 0;
uint8_t paramHelp = 0;
char *paramFEN = NULL;
uint16_t paramRandom = 0;
uint8_t paramBlind = 0;
int clockSeconds = -1;
SCL_Game game;
SCL_Board startState = SCL_BOARD_START_STATE;

int16_t makeAIMove(SCL_Board board, uint8_t *s0, uint8_t *s1, char *prom)
{
  uint8_t level = SCL_boardWhitesTurn(board) ? paramPlayerW : paramPlayerB;
  uint8_t depth = (level > 0) ? level : 1;
  uint8_t extraDepth = 4;
  uint8_t endgameDepth = 1;
  uint8_t randomness = 
    game.ply < 2 ? 1 : 0; /* in first moves increase randomness for different 
                             openings */

  uint8_t rs0, rs1;

  SCL_gameGetRepetiotionMove(&game,&rs0,&rs1);

  if (clockSeconds >= 0) // when using clock, choose AI params accordingly
  {
    if (clockSeconds <= 5)
    {
      depth = 1;
      extraDepth = 3;
      endgameDepth = 0;
    }
    else if (clockSeconds < 15)
    {
      depth = 2;
      extraDepth = 3;
    }
    else if (clockSeconds < 25)
    {
      depth = 2;
    }
    else if (clockSeconds < 5 * 60)
    {
      depth = 3;
    }
    else
    {
      depth = 3;
      extraDepth = 5;
    }
  }

  return SCL_getAIMove(board,depth,extraDepth,endgameDepth,SCL_boardEvaluateStatic,SCL_randomBetter,randomness,rs0,rs1,s0,s1,prom);
}

void printHelp(void)
{
  puts("smolchess\n");
  puts("Tiny public domain chess engine made with smallchesslib. Possible params:");
  puts("  -h   print help and exit");
  puts("  -x   start in xboard mode");
  puts("  -bN  board print format (1 - 4, 0 = don't draw board)");
  puts("  -PN  white player (0 = human, 1, 2, 3, ... AI)");
  puts("  -pN  black player");
  puts("  -aN  show AI evaluation and best move");
  puts("  -m   show possible moves and exit");
  puts("  -fS  load position from given FEN string");
  puts("  -i   show info during play (evaluation, PGN, ...)");
  puts("  -F   flip board");
  puts("  -d   draw board to a ppm image file and end");
  puts("  -rN  set pseudorandom seed");
}

int main(int argc, char **argv)
{
  paramRandom = (uint16_t) time(0) + clock();

  for (int i = 0; i < argc; ++i)
  {
    char *a = argv[i];

    if (a[0] != '-')
      continue;

    switch (a[1])
    {
      case 'h': paramHelp = 1; break;             // print help
      case 'b': paramBoard = a[2] - '0'; break;   // board format
      case 'P': paramPlayerW = a[2] - '0'; break; // player 1 (white)
      case 'p': paramPlayerB = a[2] - '0'; break; // player 2 (black)
      case 'a': paramAnalyze = (a[2] == 0) ? 0 :
                  (a[2] - '0'); break;            // show AI evaluation and move
      case 'm': paramMoves = 1; break;            // show possible moves and end
      case 'x': paramXboard = 1; break;           // start as xboard engine
      case 'f': paramFEN = a + 2; break;          // set position (FEN)
      case 'F': paramFlipBoard = 1; break;        // flip board (black down)
      case 'i': paramInfo = a[2] != '0'; break;   // show info
      case 'd': paramDraw = 1;                    // draw board to file and end
      case 'r':                                   // random seed
        {
          paramRandom = 0;

          char *s = &(a[2]);

          while (*s != 0)
          {
            paramRandom *= 10;
            paramRandom += *s - '0';
            s++;
          }
        }

        break;

      default: puts("encountered unknown parameter"); break;
    }
  }

  if (paramHelp)
  {
    printHelp();
    return 0;
  }

  SCL_randomBetterSeed(paramRandom);
 
  char string[256];

  if (paramFEN != NULL)
    SCL_boardFromFEN(startState,paramFEN);

  SCL_gameInit(&game,startState);

  SCL_SquareSet squareSet = SCL_SQUARE_SET_EMPTY;

  uint8_t move[] = {0, 0};

  if (!paramXboard)
  {
    // interactive CLI mode
  
    if (paramAnalyze != 255)
    {
      char p;

      paramPlayerW = paramAnalyze;
      paramPlayerB = paramAnalyze;

      int16_t evaluation = makeAIMove(game.board,&(move[0]),&(move[1]),&p);

      if (paramAnalyze == 0)
        evaluation = SCL_boardEvaluateStatic(game.board);

      char moveStr[5];
      moveStr[4] = 0;

      SCL_squareToString(move[0],moveStr);
      SCL_squareToString(move[1],moveStr + 2);

      printf("%lf (%d)\n",((double) evaluation) / ((double) SCL_VALUE_PAWN),evaluation);
      puts(moveStr);

      return 0;
    }
      
    if (paramMoves)
    {
      for (int i = 0; i < 64; ++i)
        if (game.board[i] != '.' && 
            SCL_pieceIsWhite(game.board[i]) == SCL_boardWhitesTurn(game.board))
        {
          SCL_SquareSet possibleMoves = SCL_SQUARE_SET_EMPTY;

          SCL_boardGetMoves(game.board,i,possibleMoves);

          SCL_SQUARE_SET_ITERATE_BEGIN(possibleMoves)
            SCL_moveToString(game.board,i,iteratedSquare,'q',string);
            printf("%s ",string);
          SCL_SQUARE_SET_ITERATE_END
        }

      return 0;
    }

    if (paramDraw)
    {
      uint8_t picture[3 * SCL_BOARD_PICTURE_WIDTH * SCL_BOARD_PICTURE_WIDTH];
      pixels = picture;

      SCL_drawBoard(game.board,putImagePixel,255,squareSet,paramFlipBoard);

      FILE *f = fopen("board.ppm","wb");

      char header[] = "P6 64 64 255\n";

      fwrite(header,sizeof(header) - 1,1,f);
      fwrite(picture,sizeof(picture),1,f);

      fclose(f);

      return 0;
    }

    char moveString[16];
    moveString[0] = 0;
    SCL_SquareSet moveHighlight = SCL_SQUARE_SET_EMPTY;

    while (1)
    {
      uint8_t moveType = 0; // 0: none, 1: player, 2: AI, 3: undo

      for (int i = 0; i < 40; ++i)
        putchar('\n');

      if (paramBoard)
        SCL_printBoard(game.board,putCharacter,moveHighlight,255,paramBoard,2,1,paramFlipBoard);

      putchar('\n');

      if (game.ply > 0)
      {
        printf(SCL_boardWhitesTurn(game.board) ? "black" : "white");
        printf(" played ");

        uint8_t s0, s1;
        char p;

        SCL_recordGetMove(game.record,game.ply - 1,&s0,&s1,&p);
        SCL_moveToString(game.board,s0,s1,p,moveString);
        printf("%s\n",moveString);
      }

      printf(SCL_boardWhitesTurn(game.board) ? "white" : "black");    
      printf(" to move\n");

      if (paramInfo)
      {
        putchar('\n');
        printf("ply number: %d\n",game.ply);
        SCL_boardToFEN(game.board,string);
        printf("FEN: %s\n",string);
        int16_t eval = SCL_boardEvaluateStatic(game.board);
        printf("board static evaluation: %lf (%d)\n",((double) eval) / ((double) SCL_VALUE_PAWN),eval);
        printf("board hash: %u\n",SCL_boardHash32(game.board));
        printf("phase: ");

        switch (SCL_boardEstimatePhase(game.board))
        {
          case SCL_PHASE_OPENING: puts("opening"); break;
          case SCL_PHASE_ENDGAME: puts("endgame"); break;
          default: puts("midgame"); break;
        }

        printf("en passant: %d\n",((game.board[SCL_BOARD_ENPASSANT_CASTLE_BYTE] & 0x0f) + 1) % 16);
        printf("50 move rule count: %d\n",game.board[SCL_BOARD_MOVE_COUNT_BYTE]);

        if (paramFEN == NULL)
        {
          printf("PGN: ");
          SCL_printPGN(game.record,putCharacter);
          putchar('\n');
        }
      } 

      if (game.state != SCL_GAME_STATE_PLAYING)
        break;
 
      uint8_t squareFrom = 0;
      uint8_t squareTo = 0;
      char movePromote = 'q';

      if (
        (SCL_boardWhitesTurn(game.board) && paramPlayerW == 0) || 
        (!SCL_boardWhitesTurn(game.board) && paramPlayerB == 0))
      {
        printf("\nmove: ");
        scanf("%s",string);

        if (stringsEqual(string,"undo",5))
          moveType = 3;
        else if (stringsEqual(string,"quit",5))
          break;
        else
        {
          squareFrom = SCL_stringToSquare(string);
          squareTo = SCL_stringToSquare(string + 2);

          uint8_t r = SCL_stringToMove(string,&squareFrom,&squareTo,&movePromote);

          if (r)
          {
            if ((game.board[squareFrom] != '.') &&
              (SCL_pieceIsWhite(game.board[squareFrom]) == SCL_boardWhitesTurn(game.board)))
              {
                SCL_boardGetMoves(game.board,squareFrom,squareSet);

                if (SCL_squareSetContains(squareSet,squareTo))
                {
                  moveType = 1;
                }
              }
          }
        }
      }
      else
      {
        makeAIMove(game.board,&squareFrom,&squareTo,&movePromote);
        moveType = 2;
      }

      if (moveType == 1 || moveType == 2)
      {
        SCL_moveToString(game.board,squareFrom,squareTo,movePromote,moveString);

        if (moveType == 1)
          SCL_gameMakeMove(&game,squareFrom,squareTo,movePromote);
        else
          SCL_gameMakeMove(&game,squareFrom,squareTo,movePromote);

        SCL_squareSetClear(moveHighlight);
        SCL_squareSetAdd(moveHighlight,squareFrom);
        SCL_squareSetAdd(moveHighlight,squareTo);
      } 
      else if (moveType == 3)
      {
        if (paramPlayerW != 0 || paramPlayerB != 0)
          SCL_gameUndoMove(&game);

        SCL_gameUndoMove(&game);
        SCL_squareSetClear(moveHighlight);
      }
    }

    putchar('\n');

    switch (game.state)
    {
      case SCL_GAME_STATE_WHITE_WIN:
        puts("white wins");
        break;

      case SCL_GAME_STATE_BLACK_WIN:
        puts("black wins");
        break;

      case SCL_GAME_STATE_DRAW_STALEMATE:
        puts("draw (stalemate)");
        break;

      case SCL_GAME_STATE_DRAW_REPETITION:
        puts("draw (repeated position)");
        break;

      case SCL_GAME_STATE_DRAW_DEAD:
        puts("draw (dead position)");
        break;

      case SCL_GAME_STATE_DRAW:
        puts("draw");
        break;

      case SCL_GAME_STATE_DRAW_50:
        puts("draw (50 move rule)");
        break;

      default: 
        puts("game over"); 
        break;
    }
  }
  else
  {
    // xboard mode
    uint8_t started = 0;
    uint8_t playingWhite = 0;

    paramPlayerW = 3;
    paramPlayerB = paramPlayerW;

    setbuf(stdout,NULL);

#if XBOARD_DEBUG
    uint8_t randomLetter;

    FILE *randFile = fopen("/dev/random","rb");            
    fread(&randomLetter,1,1,randFile);
    fclose(randFile);
    char debugFileName[] = "/home/tastyfish/git/smallchesslib/xboard_debug_x.txt";
    debugFileName[47] = randomLetter % 26 + 'A';
    debugFile = fopen(debugFileName,"w");

    fprintf(debugFile,"MY RANDOM SEED IS: '%d'\n",paramRandom);
#endif

    while (1)
    {
      char p;

      gets(string);

#if XBOARD_DEBUG
      fprintf(debugFile,"RECEIVED: '%s'\n",string);
#endif

      if (stringsEqual(string,"new",3))
      {
        SCL_boardInit(game.board);
      }
      else if (stringsEqual(string,"go",2))
      {
        if (!started)
          playingWhite = SCL_boardWhitesTurn(game.board);
          
        started = 1;
      }
      else if (stringsEqual(string,"xboard",6))
      {
      }
      else if (stringsEqual(string,"sd",2)) // search depth
      {
        paramPlayerW = string[3] - '0';

        if (paramPlayerW > 5) 
          paramPlayerW = 5;

        paramPlayerB = paramPlayerW;

#if XBOARD_DEBUG
        fprintf(debugFile,"SET SEARCH DEPTH TO %d\n",paramPlayerW);
#endif
      }
      else if (stringsEqual(string,"quit",4))
      {
        break;
      }
      else if (stringsEqual(string,"protover",8))
      {
        xboardSend("feature done=0\n");
        xboardSend("feature time=1\n");
        xboardSend("feature ping=1\n");
        xboardSend("feature setboard=1\n");
        xboardSend("feature sigint=0\n");
        xboardSend("feature sigterm=0\n");
        xboardSend("feature reuse=0\n");
        xboardSend("feature colors=0\n");
        xboardSend("feature done=1\n");
        xboardSend("feature done=1\n");
      }
      else if (stringsEqual(string,"setboard",8))
      {
        SCL_boardFromFEN(game.board,string + 9);
      }
      else if (stringsEqual(string,"time",4))
      {
        clockSeconds = 0;
        const char *s = string + 5;

        while (*s != 0)
        {
          clockSeconds = clockSeconds * 10 + *s - '0';
          s++;
        }

        clockSeconds /= 100;
      }
      else if (stringsEqual(string,"ping",4))
      {
        char reply[64] = "pong ";

        const char *s = string + 5;
        char *s2 = reply + 5;  

        while (1)
        {
          *s2 = *s;

          if (*s == 0)
            break;

          s2++;
          s++;
        }

        xboardSend(reply);
        xboardSend("\n");
      }
      else
      {
        uint8_t r = SCL_stringToMove(string,&(move[0]),&(move[1]),&p);

        if (r)
        {
          SCL_gameMakeMove(&game,move[0],move[1],p); 

          if (!started)
            playingWhite = SCL_boardWhitesTurn(game.board);

          started = 1;
        }
      }

      if (started && (playingWhite == SCL_boardWhitesTurn(game.board)))
      {
        makeAIMove(game.board,&(move[0]),&(move[1]),&p);

        SCL_moveToString(game.board,move[0],move[1],p,string);

        SCL_gameMakeMove(&game,move[0],move[1],p);

        xboardSend("move ");
        xboardSend(string);
        xboardSend("\n");
      }

#if XBOARD_DEBUG
      SCL_boardToFEN(board,string);
      debugLogBoard(board);
      fprintf(debugFile,"FEN: '%s'\n",string);
#endif
    }
#if XBOARD_DEBUG
    fclose(debugFile);
#endif
  }

  return 0;
}
